from http import HTTPStatus
from unittest import mock

import pytest
from django.http import HttpResponse
from django.test import Client, override_settings
from django.urls import reverse

from django_cloudevents.processors import InvalidEventProcessorError, event_processors
from django_cloudevents.signals import cloudevent_received

pytestmark = pytest.mark.django_db


class TestWebhookView:
    def test_options_without_request_origin(self, client: Client):
        response: HttpResponse = client.options(reverse("django_cloudevents:webhook"))

        assert response.status_code == HTTPStatus.OK
        assert "WebHook-Allowed-Origin" not in response.headers
        assert "WebHook-Allowed-Rate" not in response.headers

    @override_settings(WEBHOOK_ALLOWED_ORIGINS=["*"], WEBHOOK_ALLOWED_RATE=None)
    def test_options_with_every_allowed_origin(self, client: Client):
        response: HttpResponse = client.options(
            reverse("django_cloudevents:webhook"), headers={"WebHook-Request-Origin": "eventemitter.example.com"}
        )

        assert response.status_code == HTTPStatus.OK
        assert response.headers["WebHook-Allowed-Origin"] == "*"
        assert "WebHook-Allowed-Rate" not in response.headers

    @override_settings(WEBHOOK_ALLOWED_ORIGINS=["eventemitter.example.com"], WEBHOOK_ALLOWED_RATE=None)
    def test_options_with_allowed_origin(self, client: Client):
        response: HttpResponse = client.options(
            reverse("django_cloudevents:webhook"), headers={"WebHook-Request-Origin": "eventemitter.example.com"}
        )

        assert response.status_code == HTTPStatus.OK
        assert response.headers["WebHook-Allowed-Origin"] == "eventemitter.example.com"
        assert "WebHook-Allowed-Rate" not in response.headers

    @override_settings(WEBHOOK_ALLOWED_ORIGINS=["eventemitter.example.com"], WEBHOOK_ALLOWED_RATE=None)
    def test_options_with_allowed_origin_and_rate(self, client: Client):
        response: HttpResponse = client.options(
            reverse("django_cloudevents:webhook"),
            headers={
                "WebHook-Request-Origin": "eventemitter.example.com",
                "WebHook-Request-Rate": "100",
            },
        )

        assert response.status_code == HTTPStatus.OK
        assert response.headers["WebHook-Allowed-Origin"] == "eventemitter.example.com"
        assert response.headers["WebHook-Allowed-Rate"] == "100"

    @override_settings(
        WEBHOOK_ALLOWED_ORIGINS=["eventemitter.example.com"],
        WEBHOOK_ALLOWED_RATE=50,
    )
    def test_options_with_allowed_origin_and_custom_rate(self, client: Client):
        response: HttpResponse = client.options(
            reverse("django_cloudevents:webhook"),
            headers={
                "WebHook-Request-Origin": "eventemitter.example.com",
                "WebHook-Request-Rate": "100",
            },
        )

        assert response.status_code == HTTPStatus.OK
        assert response.headers["WebHook-Allowed-Origin"] == "eventemitter.example.com"
        assert response.headers["WebHook-Allowed-Rate"] == "50"

    @override_settings(
        WEBHOOK_ALLOWED_ORIGINS=["eventemitter.example.com"],
        WEBHOOK_ALLOWED_RATE="*",
    )
    def test_options_with_allowed_origin_and_unlimited_rate(self, client: Client):
        response: HttpResponse = client.options(
            reverse("django_cloudevents:webhook"),
            headers={
                "WebHook-Request-Origin": "eventemitter.example.com",
                "WebHook-Request-Rate": "100",
            },
        )

        assert response.status_code == HTTPStatus.OK
        assert response.headers["WebHook-Allowed-Origin"] == "eventemitter.example.com"
        assert response.headers["WebHook-Allowed-Rate"] == "*"

    @override_settings(
        WEBHOOK_ALLOWED_ORIGINS=["eventemitter.example.com"],
        WEBHOOK_ALLOWED_RATE="*",
    )
    def test_options_with_denied_origin(self, client: Client):
        response: HttpResponse = client.options(
            reverse("django_cloudevents:webhook"),
            headers={
                "WebHook-Request-Origin": "denied.example.com",
                "WebHook-Request-Rate": "100",
            },
        )

        assert response.status_code == HTTPStatus.OK
        assert "WebHook-Allowed-Origin" not in response.headers
        assert "WebHook-Allowed-Rate" not in response.headers

    def test_post_calls_signals(self, cloudevent, client: Client):
        mock_signal = mock.Mock()

        cloudevent_received.connect(mock_signal)
        try:
            client.post(reverse("django_cloudevents:webhook"), data=cloudevent, content_type="application/json")

            mock_signal.assert_called_once()
        finally:
            cloudevent_received.disconnect(mock_signal)

    def test_post_no_processors(self, cloudevent, client: Client):
        with mock.patch.object(event_processors, "create_connection", side_effect=InvalidEventProcessorError):
            response: HttpResponse = client.post(
                reverse("django_cloudevents:webhook"), data=cloudevent, content_type="application/json"
            )

        assert response.status_code == HTTPStatus.UNSUPPORTED_MEDIA_TYPE

    @mock.patch("django_cloudevents.views.event_processors")
    def test_post_processor_with_response(self, mock_handler, cloudevent, client: Client):
        mock_processor = mock.AsyncMock()
        expected = HttpResponse("", status=HTTPStatus.OK)
        mock_processor.aprocess_event.return_value = expected
        mock_handler.__getitem__.return_value = mock_processor

        response: HttpResponse = client.post(
            reverse("django_cloudevents:webhook"), data=cloudevent, content_type="application/json"
        )

        assert response == expected

    @mock.patch("django_cloudevents.views.event_processors")
    def test_post_processor_without_response_returns_accepted(self, mock_handler, cloudevent, client: Client):
        mock_processor = mock.AsyncMock()
        mock_processor.aprocess_event.return_value = None
        mock_handler.__getitem__.return_value = mock_processor

        response: HttpResponse = client.post(
            reverse("django_cloudevents:webhook"), data=cloudevent, content_type="application/json"
        )

        assert response.status_code == HTTPStatus.ACCEPTED



import pytest


@pytest.fixture
def cloudevent():
    return {
        "specversion": "1.0",
        "type": "com.github.pull_request.opened",
        "source": "https://github.com/cloudevents/spec/pull",
        "subject": "123",
        "id": "A234-1234-1234",
        "time": "2018-04-05T17:31:00Z",
        "comexampleextension1": "value",
        "comexampleothervalue": 5,
        "datacontenttype": "text/xml",
        "data": '<much wow="xml"/>',
    }



from http import HTTPStatus

from cloudevents.abstract import CloudEvent
from cloudevents.conversion import to_structured
from django.http import HttpResponse

from django_cloudevents.processors import SyncEventProcessor


class EchoEventProcessor(SyncEventProcessor):
    def __init__(self, *, status_code: HTTPStatus = HTTPStatus.ACCEPTED):
        self.status_code = status_code

    def process_event(self, cloudevent: CloudEvent) -> HttpResponse:
        headers, data = to_structured(cloudevent)
        return HttpResponse(data, status=self.status_code, headers=headers)



from http import HTTPStatus
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

SECRET_KEY = "django-insecure-txy9b_=khny3ax$up%i&)&axs!inx7dy1b0nu26g!q-z$b)f-f"  # noqa: S105

DEBUG = True

ALLOWED_HOSTS = []

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_cloudevents",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "testproj.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "testproj.wsgi.application"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    },
    "dummy": {
        "ENGINE": "django.db.backends.dummy",
    },
}

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
    },
    "dummy": {
        "BACKEND": "django.core.cache.backends.dummy.DummyCache",
    },
}

STORAGES = {
    "default": {
        "BACKEND": "django.core.files.storage.InMemoryStorage",
    },
    "staticfiles": {
        "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
    },
}

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


LANGUAGE_CODE = "en-us"
TIME_ZONE = "UTC"
USE_I18N = True
USE_TZ = True

STATIC_URL = "static/"

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

WEBHOOK_ALLOWED_ORIGINS = ["*"]
WEBHOOK_ALLOWED_RATE = "*"
CLOUDEVENT_PROCESSORS = {
    "order": {
        "BACKEND": "testproj.processors.EchoEventProcessor",
        "OPTIONS": {
            "status_code": HTTPStatus.OK,
        },
    },
    "stock": {
        "BACKEND": "django_cloudevents.processors.AcceptEventProcessor",
    },
}



import pytest
from cloudevents.conversion import from_dict
from cloudevents.http import CloudEvent
from django.conf import settings

from django_cloudevents.processors import AcceptEventProcessor, EventHandler, InvalidEventProcessorError


class TestAcceptEventProcessor:
    def test_process_event(self, cloudevent):
        given = from_dict(CloudEvent, cloudevent)
        processor = AcceptEventProcessor()

        assert processor.process_event(given) is None

    @pytest.mark.asyncio
    async def test_aprocess_event(self, cloudevent):
        given = from_dict(CloudEvent, cloudevent)
        processor = AcceptEventProcessor()

        assert await processor.aprocess_event(given) is None


class TestEventHandler:
    def test_with_custom_settings(self):
        handler = EventHandler(
            {
                "test": {
                    "BACKEND": "django_cloudevents.processors.AcceptEventProcessor",
                }
            }
        )
        items = set(handler)

        assert items == {"test"}

    def test_with_default_settings(self):
        handler = EventHandler()
        items = set(handler)

        assert items == set(settings.CLOUDEVENT_PROCESSORS)

    def test_get_existing_item(self):
        handler = EventHandler(
            {
                "test": {
                    "BACKEND": "django_cloudevents.processors.AcceptEventProcessor",
                }
            }
        )

        got = handler["test"]

        assert isinstance(got, AcceptEventProcessor)

    def test_get_missing_item(self):
        handler = EventHandler(
            {
                "test": {
                    "BACKEND": "django_cloudevents.processors.AcceptEventProcessor",
                }
            }
        )

        with pytest.raises(InvalidEventProcessorError):
            handler["missing"]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Protocol, TypedDict

from asgiref.sync import async_to_sync, sync_to_async
from cloudevents.abstract import CloudEvent
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse
from django.utils.connection import BaseConnectionHandler
from django.utils.module_loading import import_string

if TYPE_CHECKING:
    import re
    from collections.abc import Mapping

    from cloudevents.abstract import CloudEvent
    from django.http import HttpResponse


class EventProcessor(Protocol):
    def process_event(self, cloudevent: CloudEvent) -> HttpResponse | None:
        pass

    async def aprocess_event(self, cloudevent: CloudEvent) -> HttpResponse | None:
        pass


class SyncEventProcessor(ABC):
    @abstractmethod
    def process_event(self, cloudevent: CloudEvent) -> HttpResponse | None:
        pass

    async def aprocess_event(self, cloudevent: CloudEvent) -> HttpResponse | None:
        return await sync_to_async(self.process_event)(cloudevent)


class AsyncEventProcessor(ABC):
    def process_event(self, cloudevent: CloudEvent) -> HttpResponse | None:
        return async_to_sync(self.aprocess_event)(cloudevent)

    @abstractmethod
    async def aprocess_event(self, cloudevent: CloudEvent) -> HttpResponse | None:
        pass


class AcceptEventProcessor(AsyncEventProcessor):
    def process_event(self, cloudevent: CloudEvent) -> HttpResponse | None:  # noqa: ARG002
        return None

    async def aprocess_event(self, cloudevent: CloudEvent) -> HttpResponse | None:  # noqa: ARG002
        return None


class InvalidEventProcessorError(ImproperlyConfigured):
    pass


class EventProcessorConfig(TypedDict):
    BACKEND: str
    SUBJECT: re.Pattern
    OPTIONS: Mapping[str, Any]


class EventHandler(BaseConnectionHandler):
    settings_name = "CLOUDEVENT_PROCESSORS"
    exception_class = InvalidEventProcessorError

    def create_connection(self, alias: str) -> EventProcessor:
        params: EventProcessorConfig = self.settings[alias]
        backend: str = params["BACKEND"]
        options: Mapping[str, Any] = params.get("OPTIONS", {})

        try:
            factory = import_string(backend)
        except ImportError as e:
            msg = f"Could not find backend {backend!r}: {e}"
            raise InvalidEventProcessorError(msg) from e
        else:
            return factory(**options)


event_processors = EventHandler()



from __future__ import annotations

from http import HTTPStatus
from typing import TYPE_CHECKING, Any, ClassVar

from cloudevents.http import from_http
from django.http import HttpResponse
from django.http.request import validate_host
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from .conf import settings
from .processors import InvalidEventProcessorError, event_processors
from .signals import cloudevent_received

if TYPE_CHECKING:
    from django.http import HttpRequest


class WebhookView(View):
    http_method_names: ClassVar[list[str]] = ["post", "options"]

    async def post(self, request: HttpRequest) -> HttpResponse:
        cloudevent = from_http(request.headers, request.body)
        await cloudevent_received.asend(None, cloudevent=cloudevent)

        try:
            event_processor = event_processors[cloudevent["type"]]
        except InvalidEventProcessorError:
            return HttpResponse(status=HTTPStatus.UNSUPPORTED_MEDIA_TYPE)
        else:
            response = await event_processor.aprocess_event(cloudevent)
            if not response:
                return HttpResponse(status=HTTPStatus.ACCEPTED)
            return response

    async def options(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        response = await super().options(request, *args, **kwargs)

        if "WebHook-Request-Origin" in request.headers and validate_host(
            request.headers["WebHook-Request-Origin"],
            settings.webhook_allowed_origins,
        ):
            response["WebHook-Allowed-Origin"] = (
                "*" if settings.webhook_allow_all_origins else request.headers["WebHook-Request-Origin"]
            )

            if settings.webhook_allowed_rate:
                response["WebHook-Allowed-Rate"] = str(settings.webhook_allowed_rate)
            elif "WebHook-Request-Rate" in request.headers:
                response["WebHook-Allowed-Rate"] = request.headers["WebHook-Request-Rate"]
        return response

    @method_decorator(csrf_exempt)
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        return super().dispatch(request, *args, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from django.conf import settings as django_settings
from django.core.signals import setting_changed

if TYPE_CHECKING:
    from collections.abc import Sequence


class Settings:
    @property
    def webhook_allowed_origins(self) -> Sequence[str]:
        return getattr(django_settings, "WEBHOOK_ALLOWED_ORIGINS", ["*"])

    @property
    def webhook_allow_all_origins(self) -> bool:
        return self.webhook_allowed_origins == ["*"]

    @property
    def webhook_allowed_rate(self) -> int | Literal["*"] | None:
        return getattr(django_settings, "WEBHOOK_ALLOWED_RATE", None)


settings = Settings()


def reload_settings(*args, **kwargs):  # noqa: ARG001
    setting = kwargs["setting"]
    if setting == "CLOUDEVENTS":
        settings.settings = getattr(django_settings, "CLOUDEVENTS", {})


setting_changed.connect(reload_settings)



import django
from asgiref.sync import sync_to_async
from django.dispatch import Signal as _DjangoSignal

if django.VERSION >= (5, 0):
    Signal = _DjangoSignal
else:

    class Signal(_DjangoSignal):  # type: ignore[no-redef]
        async def asend(self, *args, **kwargs):
            return await sync_to_async(self.send)(*args, **kwargs)


__all__ = [
    "Signal",
]
